!
! Copyright (c) 1983-2000,2013 TAPP Consortium
! Copyright (c) 2000-2017 Yoshihide Yoshimoto
!
module util
contains
  real(kind=8) function cellvol(aa)
    !
    !       calculates the cell volume of aa
    !
    implicit none
    real(kind=8),intent(in):: aa(3,3)
    !
    real(kind=8) v
    !
    v=aa(1,1)*(aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3)) &
         +aa(2,1)*(aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3)) &
         +aa(3,1)*(aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3))
    cellvol=abs(v)
    return
  end function cellvol
  !
  subroutine outer_product(a,b,c)
    !
    !       calculates the outer product c = a \cross b
    !
    implicit none
    real(kind=8),intent(in):: a(3),b(3)
    real(kind=8),intent(out):: c(3)
    !        
    c(1) = a(2)*b(3)-a(3)*b(2)
    c(2) = a(3)*b(1)-a(1)*b(3)
    c(3) = a(1)*b(2)-a(2)*b(1)
    return
  end subroutine outer_product
  !
  integer function suplatp(aa,r)
    implicit none
    real(kind=8),intent(in)::aa(3,3),r
    !
    real(kind=8),parameter:: pi = 3.1415926535897932385d0
    real(kind=8):: x(3),d
    !
    ! calculate cell diameter d.
    !
    x(:) = aa(:,1) + aa(:,2) + aa(:,3)
    d = sqrt(dot_product(x,x))
    !
    x(:) = aa(:,1) + aa(:,2) - aa(:,3)
    d = max(d, sqrt(dot_product(x,x)))
    !
    x(:) = aa(:,1) - aa(:,2) + aa(:,3)
    d = max(d, sqrt(dot_product(x,x)))
    !
    x(:) = aa(:,1) - aa(:,2) - aa(:,3)
    d = max(d, sqrt(dot_product(x,x)))
    !
    suplatp = int((4.0d0*pi/3.0d0)*(r+d)**3/cellvol(aa))
    return
  end function suplatp
  !
  subroutine LatticeInterval(aa,h)
    implicit none
    real(kind=8),intent(in):: aa(3,3)
    real(kind=8),intent(out):: h(3)
    !
    real(kind=8):: hh(3,3),htmp
    !
    ! hh(*,1) = aa(*,2) \times aa(*,3)
    ! hh(*,2) = aa(*,3) \times aa(*,1)
    ! hh(*,3) = aa(*,1) \times aa(*,2)
    !
    call outer_product(aa(:,2),aa(:,3),hh(:,1))
    call outer_product(aa(:,3),aa(:,1),hh(:,2))
    call outer_product(aa(:,1),aa(:,2),hh(:,3))
    !
    ! normalize hh(*,1),hh(*,2),hh(*,3)
    !
    htmp=sqrt(dot_product(hh(:,1),hh(:,1)))
    hh(:,1)=hh(:,1)/htmp
    htmp=sqrt(dot_product(hh(:,2),hh(:,2)))
    hh(:,2)=hh(:,2)/htmp
    htmp=sqrt(dot_product(hh(:,3),hh(:,3)))
    hh(:,3)=hh(:,3)/htmp
    !
    ! h(1): interval of the plane (1,0,0) etc.
    !
    h(1)=abs(dot_product(aa(:,1),hh(:,1)))
    h(2)=abs(dot_product(aa(:,2),hh(:,2)))
    h(3)=abs(dot_product(aa(:,3),hh(:,3)))
    return
  end subroutine LatticeInterval
  !
  subroutine fctck(n,m2,m3,m5,info)
    implicit none
    integer,intent(in):: n
    integer,intent(out):: m2, m3, m5, info
    !
    integer:: i
    !
    i=n
    m2 = 0
    m3 = 0
    m5 = 0
    info = 0
    !
    do while (i .ne. 1)
       if (mod(i,2) .eq. 0) then
          m2 = m2 + 1
          i = i / 2
       else if (mod(i,3) .eq. 0) then
          m3 = m3 + 1
          i = i / 3
       else if (mod(i,5) .eq. 0) then
          m5 = m5 + 1
          i = i / 5
       else
          info = 1
          exit
       end if
    end do
    return
  end subroutine fctck
!
  integer function algn235(inr,gctr)
    implicit none
    integer,intent(in):: inr,gctr
!
    integer:: nr,m2,m3,m5,info
    integer:: t2, t3, t5
!
    call fctck(gctr,t2,t3,t5,info)
    if (info.eq.1) then
       write(6,*) 'algn235: gctr is not a multiple of 2,3,5'
       stop
    end if
    nr = inr
    call fctck(nr,m2,m3,m5,info)
    do while ( (info .eq. 1) .or. (m2.lt.t2) &
         .or. (m3.lt.t3) .or. (m5.lt.t5) )
       nr = nr + 1
       call fctck(nr,m2,m3,m5,info)
    end do
    algn235 = nr
    return
  end function algn235
!
  logical function normconsp(nl,nr,maxr,ifcr)
    implicit none
    integer,intent(in):: nl, nr(nl), maxr, ifcr(maxr,nl)
    integer:: il, ir
    logical:: cnd

    cnd = .true.
!
! These loops are very short.
!
!POPTION NOPARALLEL
    do il=1,nl
!POPTION NOPARALLEL
      do ir=1,nr(il)
         if (ifcr(ir,il) .gt. 0) then
            cnd = .false.
         end if
      end do
    end do
    normconsp = cnd
    return
  end function normconsp

  logical function hasinvsym(ng,nnp,rg,pg,srg)
    implicit none
    integer,intent(in):: ng,nnp,rg(3,3,ng), pg(3,ng), srg(ng)

    integer:: ig
    logical:: retval

    retval = .false.
    do ig=1,ng
       if( (rg(1,1,ig).eq.-1) .and. (rg(2,1,ig).eq.0) &
                              .and. (rg(3,1,ig).eq.0) .and. &
           (rg(1,2,ig).eq.0) .and. (rg(2,2,ig).eq.-1) &
                             .and. (rg(3,2,ig).eq.0) .and. &
           (rg(1,3,ig).eq.0) .and. (rg(2,3,ig).eq.0) &
                             .and. (rg(3,3,ig).eq.-1) .and. &
           (mod(pg(1,ig),nnp).eq.0) .and. &
           (mod(pg(2,ig),nnp).eq.0) .and. &
           (mod(pg(3,ig),nnp).eq.0) .and. srg(ig).eq.1 ) then
          retval = .true.
       end if
    end do
    hasinvsym = retval
    return
  end function hasinvsym

  subroutine mkstrfac(nsi,asi,nkm,kg,strfac)
    implicit none
    integer,intent(in):: nsi, nkm, kg(3,nkm)
    real(kind=8),intent(in):: asi(3,nsi)
    complex(kind=8),intent(out):: strfac(nkm,nsi)

    real(kind=8),parameter:: pi = 3.1415926535897932385d0
    real(kind=8):: ph
    integer:: it,ig

    do it=1,nsi
       do ig=1,nkm
          ph=2.0d0*pi*(asi(1,it)*kg(1,ig) &
               +asi(2,it)*kg(2,ig)+asi(3,it)*kg(3,ig))
          strfac(ig,it) = dcmplx(cos(ph),-sin(ph))
       end do
    end do
    return
  end subroutine mkstrfac

  real(kind=8) function difflpt(nvxyz, old, new)
    implicit none
    integer,intent(in):: nvxyz
    complex(kind=8),intent(in):: old(nvxyz), new(nvxyz)

    integer:: i
    complex(kind=8):: diff
    real(kind=8):: r2
      
    r2 = 0.0d0
    do i=1,nvxyz
       diff =  new(i) - old(i)
       r2 = r2 + dble(diff)*dble(diff) + aimag(diff)*aimag(diff)
    end do
!
!   \frac{1}{\Omega} \int_{\Omega} |V_{new}(r) - V_{old}(r)|^2 dr
! = \sum_{G} ( \frac{1}{\Omega} |V_{new}(G) - V_{old}(G)| )^2
!
    difflpt = r2

    return
  end function difflpt

  !  
  ! from Research Institute for Mathematical Sciences, Kyoto University
  ! Ko^kyu^roku vol. 1032, 21-45 (1998)
  ! Satoru Tanaka, Shuya Kanagawa
  ! pseudo randoum number B using Schrage technique
  !
  real(kind=8) function ran0(state)
    implicit none
    integer:: state

    integer,parameter:: a = 16807 ! = 7^5
    integer,parameter:: m = 2147483647 ! = 2^31 - 1
    integer,parameter:: r = 2836 ! = m/a
    integer,parameter:: q = 127773 ! = mod(m,a)
    integer:: tmp

    if (state.eq.0) then
       write(6,*) 'ran0: state = 0'
       stop
    end if

    tmp = state/q
    state = a*(state - q*tmp) - r*(state/q)
    if (state.lt.0) then
       state = state + m
    end if

    ran0 = dble(1.0d0)/dble(m) * state
    return
  end function ran0

  ! generate gaussian distribution G, from uniform distribution U
  ! E[U] = 1/2, V[U] = 1/12
  ! use  central limit theorem.
  ! ( 1/N \sum U_i - E[U] ) / ( V[U] / \sqrt N ) -> N(0,1)
  ! G = N(0,1)*sd = N(0,sd)
  real(kind=8) function rndgauss(sd,idum)
    implicit none
    real(kind=8),intent(in):: sd
    integer:: idum
    integer,parameter:: rnk = 12
    integer:: i
    real(kind=8):: g

    g = 0.0d0
    do i=1,rnk
       g = g + ran0(idum)
    end do
    g = g/rnk
    g = sqrt(dble(rnk))*(g - 0.5d0)*sqrt(12.0d0)
    rndgauss = g*sd
    return
  end function rndgauss

  integer function grtcomdiv(a,b)
    implicit none
    integer,intent(in):: a,b
    integer:: p,q,r

    if (a .eq. 0) then
       grtcomdiv = b
       return
    else if (b .eq. 0) then
       grtcomdiv = a
       return
    end if

    if (a .ge. b) then
       p = a
       q = b
    else
       p = b
       q = a
    end if

    do while (.true.)
       r = mod(p,q)
       if (r.eq.0) exit
       p = q
       q = r
    end do
    grtcomdiv = q
    return
  end function grtcomdiv

  subroutine grtcomtrns(ng,pg,nnp,gctr)
    implicit none
    integer,intent(in):: ng, pg(3,ng), nnp
    integer,intent(out):: gctr(3)

    integer:: i,ig
    do i=1,3
       gctr(i) = nnp
       do ig=1,ng
          gctr(i) = grtcomdiv(gctr(i),modulo(pg(i,ig),nnp))
       end do
       gctr(i) = nnp/gctr(i)
    end do
    return
  end subroutine grtcomtrns

  !
  ! cdm: real symmetric matrix
  ! AA = exp(cdm)*AA0
  ! cdm = V*D*inv(V)
  ! exp(cdm) = V*exp(D)*inv(V)
  !
  subroutine cdm2aa(aa0,cdm,aa)
    implicit none
    real(kind=8),intent(in):: aa0(3,3),cdm(6)
    real(kind=8),intent(out):: aa(3,3)

    integer,parameter:: lwork = 15
    real(kind=8):: tmp1(3,3),tmp2(3),work(lwork),tmp3(3,3),tmp4(3,3)
    integer:: info,i,j,k

    tmp1(1,1) = cdm(1)
    tmp1(2,2) = cdm(2)
    tmp1(3,3) = cdm(3)
    tmp1(2,3) = cdm(4)
    tmp1(3,2) = cdm(4)
    tmp1(3,1) = cdm(5)
    tmp1(1,3) = cdm(5)
    tmp1(1,2) = cdm(6)
    tmp1(2,1) = cdm(6)

    call dsyev('V','L',3,tmp1,3,tmp2,work,lwork,info)
    if (info.ne.0) then
       write(6,*) 'cdm2aa:dsyev info = ', info
       stop
    end if

    tmp4 = 0.0d0
    tmp4(1,1) = exp(tmp2(1))
    tmp4(2,2) = exp(tmp2(2))
    tmp4(3,3) = exp(tmp2(3))

    tmp3 = 0.0d0
    do k=1,3
       do j=1,3
          do i=1,3
             tmp3(i,k) = tmp3(i,k) + tmp4(i,j)*tmp1(k,j)
          end do
       end do
    end do

    tmp4 = 0.0d0
    do k=1,3
       do j=1,3
          do i=1,3
             tmp4(i,k) = tmp4(i,k) + tmp1(i,j)*tmp3(j,k)
          end do
       end do
    end do

    aa = 0.0d0
    do k=1,3
       do j=1,3
          do i=1,3
             aa(i,k) = aa(i,k) + tmp4(i,j)*aa0(j,k)
          end do
       end do
    end do

    return
  end subroutine cdm2aa

  subroutine getylb(nkm,sk,kg,bb,ylm,lmax,ncomp)
    implicit none
    integer,intent(in):: nkm,lmax,ncomp
    integer,intent(in):: kg(3,nkm)
    real(kind=8),intent(in):: bb(3,3),sk(3)
    complex(kind=8):: ylm(ncomp*nkm,ncomp*lmax**2)

    if (ncomp.eq.1) then
       call getylb_one(nkm,sk,kg,bb,ylm,lmax)
    else if (ncomp.eq.2) then
       call getylb_two(nkm,sk,kg,bb,ylm,lmax)
    end if
    return
  end subroutine getylb

!*********************************************************************
!  ACCORDING TO SLATER DEFINITION
!                SEE "The Calculation of Molecular Orbitals"
!  YLM(NG2Q,I) FOR YLM_L^M
!     1:(0, 0),2:(1,-1),3:(1, 0),4:(1, 1)
!     5:(2,-2),6:(2,-1),7:(2, 0),8:(2, 1),9:(2,2)
!          1992-07-17  Jun Yamauchi
!********************************************************************
  subroutine getylb_one(nkm,sk,kg,bb,ylm,lmax)
    implicit none
    integer,intent(in):: nkm,kg(3,nkm),lmax
    real(kind=8),intent(in):: bb(3,3),sk(3)
    complex(kind=8):: ylm(nkm,lmax**2)

    real(kind=8),parameter:: pi = 3.14159265358979323846d0
    real(kind=8):: s1,s2,s3,s0,f00,f10,f20,f21,f22,r,r2
    real(kind=8):: f30,f31,f32,f33,r3
    integer:: lmax2,ista,il,ig

    s1=bb(1,1)*(sk(1)+kg(1,1))+bb(1,2)*(sk(2)+kg(2,1)) &
         +bb(1,3)*(sk(3)+kg(3,1))
    s2=bb(2,1)*(sk(1)+kg(1,1))+bb(2,2)*(sk(2)+kg(2,1)) &
         +bb(2,3)*(sk(3)+kg(3,1))
    s3=bb(3,1)*(sk(1)+kg(1,1))+bb(3,2)*(sk(2)+kg(2,1)) &
         +bb(3,3)*(sk(3)+kg(3,1))
    s0=s1*s1+s2*s2+s3*s3
    lmax2=lmax*lmax

    f00=sqrt( 1.d0 / ( 4.d0*pi) )
    f10=sqrt( 3.d0 / ( 4.d0*pi) )
    f20=sqrt( 15.d0/pi )*0.5d0
    f21=sqrt( 5.0d0/pi )*0.25d0
    f22=f20*0.5d0

    !
    ! Imported from TAPP-3.3.0
    ! Originally developed by T. Kosugi 2007/6/22
    !
    f30=sqrt(35.d0/(2.d0*pi))*0.25d0
    f31=sqrt(105.d0/pi)
    f32=sqrt(21.d0/(2.d0*pi))*0.25d0
    f33=sqrt(7.d0/pi)*0.25d0

    if(s0.eq.0.d0) then
       ylm(1,1)= f00
       do il=2,lmax2
          ylm(1,il)=0.d0
       end do
       ista=2
    else
       ista=1
    endif
    if (lmax.eq.1) then
       do ig=ista,nkm
          ylm(ig,1)= f00
       end do
    else if (lmax.eq.2) then
       do ig=ista,nkm
          s1=bb(1,1)*(sk(1)+kg(1,ig))+bb(1,2)*(sk(2)+kg(2,ig)) &
               +bb(1,3)*(sk(3)+kg(3,ig))
          s2=bb(2,1)*(sk(1)+kg(1,ig))+bb(2,2)*(sk(2)+kg(2,ig)) &
               +bb(2,3)*(sk(3)+kg(3,ig))
          s3=bb(3,1)*(sk(1)+kg(1,ig))+bb(3,2)*(sk(2)+kg(2,ig)) &
               +bb(3,3)*(sk(3)+kg(3,ig))
          s0=s1*s1+s2*s2+s3*s3
          r=sqrt(s0)
          ylm(ig,1)= f00
          ylm(ig,2)=-f10*s2/r
          ylm(ig,3)= f10*s3/r
          ylm(ig,4)=-f10*s1/r
       end do
    else if (lmax.eq.3) then
       do ig=ista,nkm
          s1=bb(1,1)*(sk(1)+kg(1,ig))+bb(1,2)*(sk(2)+kg(2,ig)) &
               +bb(1,3)*(sk(3)+kg(3,ig))
          s2=bb(2,1)*(sk(1)+kg(1,ig))+bb(2,2)*(sk(2)+kg(2,ig)) &
               +bb(2,3)*(sk(3)+kg(3,ig))
          s3=bb(3,1)*(sk(1)+kg(1,ig))+bb(3,2)*(sk(2)+kg(2,ig)) &
               +bb(3,3)*(sk(3)+kg(3,ig))
          s0=s1*s1+s2*s2+s3*s3
          r=sqrt(s0)
          r2=r*r
          ylm(ig,1)= f00
          ylm(ig,2)=-f10*s2/r
          ylm(ig,3)= f10*s3/r
          ylm(ig,4)=-f10*s1/r
          ylm(ig,5)= f20*s1*s2/r2
          ylm(ig,6)=-f20*s2*s3/r2
          ylm(ig,7)= f21*(3.0d0*s3**2/r2-1.0d0)
          ylm(ig,8)=-f20*s3*s1/r2
          ylm(ig,9)= f22*(s1**2-s2**2)/r2
       end do
    else if (lmax.eq.4) then
       do ig=ista,nkm
          s1=bb(1,1)*(sk(1)+kg(1,ig))+bb(1,2)*(sk(2)+kg(2,ig)) &
               +bb(1,3)*(sk(3)+kg(3,ig))
          s2=bb(2,1)*(sk(1)+kg(1,ig))+bb(2,2)*(sk(2)+kg(2,ig)) &
               +bb(2,3)*(sk(3)+kg(3,ig))
          s3=bb(3,1)*(sk(1)+kg(1,ig))+bb(3,2)*(sk(2)+kg(2,ig)) &
               +bb(3,3)*(sk(3)+kg(3,ig))
          s0=s1*s1+s2*s2+s3*s3
          r=sqrt(s0)
          r2=r*r
          r3=r2*r
          !
          ! Imported from TAPP-3.3.0
          ! Originally developed by T. Kosugi 2007/6/22
          ! Modified by J. Yamauchi 2008/3/06
          !
          ylm(ig,1)= f00
          ylm(ig,2)=-f10*s2/r
          ylm(ig,3)= f10*s3/r
          ylm(ig,4)=-f10*s1/r
          ylm(ig,5)= f20*s1*s2/r2
          ylm(ig,6)=-f20*s2*s3/r2
          ylm(ig,7)= f21*(3.0d0*s3**2/r2-1.0d0)
          ylm(ig,8)=-f20*s3*s1/r2
          ylm(ig,9)= f22*(s1**2-s2**2)/r2
          ylm(ig,10) =-f30*s2*(3.d0*s1*s1-s2*s2)/r3
          ylm(ig,11) = f31*0.5d0*s1*s2*s3/r3
          ylm(ig,12) =-f32*s2/r*(5.d0*s3*s3/r2-1.d0)
          ylm(ig,13) = f33*s3/r*(5.d0*s3*s3/r2-3.d0)
          ylm(ig,14) =-f32*s1/r*(5.d0*s3*s3/r2-1.d0)
          ylm(ig,15) = f31*0.25d0*s3*(s1*s1-s2*s2)/r3
          ylm(ig,16) =-f30*s1*(s1*s1-3.d0*s2*s2)/r3
       end do
    else
       write(6,*)'LMAX must be <= 4.  LMAX=',lmax
       stop 'GETYLB_one'
    end if
    return
  end subroutine getylb_one

  subroutine getylb_two(nkm,sk,kg,bb,ylpmj,lmax)
    implicit none
    integer,intent(in):: nkm,kg(3,nkm),lmax
    real(kind=8),intent(in):: bb(3,3),sk(3)
    complex(kind=8):: ylpmj(2,nkm,2*lmax**2)

    real(kind=8),parameter:: pi = 3.14159265358979323846d0

    real(kind=8):: x,y,z,r
    integer:: ista,ig
    complex(kind=8):: iypx, iymx

    real(kind=8),parameter:: sq2 = sqrt(2.0d0), &
         sq3 = sqrt(3.0d0), sq5 = sqrt(5.0d0), &
         sq6 = sqrt(6.0d0), sq8 = sqrt(8.0d0), &
         sq15 = sqrt(15.0d0), sq30 = sqrt(30.0d0), &
         sq35 = sqrt(35.0d0)

    x = bb(1,1)*(sk(1)+kg(1,1))+bb(1,2)*(sk(2)+kg(2,1)) &
         +bb(1,3)*(sk(3)+kg(3,1))
    y = bb(2,1)*(sk(1)+kg(1,1))+bb(2,2)*(sk(2)+kg(2,1)) &
         +bb(2,3)*(sk(3)+kg(3,1))
    z = bb(3,1)*(sk(1)+kg(1,1))+bb(3,2)*(sk(2)+kg(2,1)) &
         +bb(3,3)*(sk(3)+kg(3,1))
    r = x*x + y*y + z*z

    if(r.eq.0.d0) then
       ylpmj(:,1,:) = 0
       ylpmj(1,1,1) = 0
       ylpmj(2,1,1) = 1
       ylpmj(1,1,2) = 1
       ylpmj(2,1,2) = 0
       ista=2
    else
       ista=1
    endif

    select case(lmax)
    case(1)
       do ig=ista,nkm
          x = bb(1,1)*(sk(1)+kg(1,ig))+bb(1,2)*(sk(2)+kg(2,ig)) &
               +bb(1,3)*(sk(3)+kg(3,ig))
          y = bb(2,1)*(sk(1)+kg(1,ig))+bb(2,2)*(sk(2)+kg(2,ig)) &
               +bb(2,3)*(sk(3)+kg(3,ig))
          z = bb(3,1)*(sk(1)+kg(1,ig))+bb(3,2)*(sk(2)+kg(2,ig)) &
               +bb(3,3)*(sk(3)+kg(3,ig))
          r = sqrt(x*x + y*y + z*z)
          x = x/r
          y = y/r
          z = z/r
          iypx = dcmplx(x,y)
          iymx = dcmplx(-x,y)
          ylpmj(1,ig,1) = 0
          ylpmj(2,ig,1) = 1
          ylpmj(1,ig,2) = 1
          ylpmj(2,ig,2) = 0
       end do
    case(2)
       do ig=ista,nkm
          x = bb(1,1)*(sk(1)+kg(1,ig))+bb(1,2)*(sk(2)+kg(2,ig)) &
               +bb(1,3)*(sk(3)+kg(3,ig))
          y = bb(2,1)*(sk(1)+kg(1,ig))+bb(2,2)*(sk(2)+kg(2,ig)) &
               +bb(2,3)*(sk(3)+kg(3,ig))
          z = bb(3,1)*(sk(1)+kg(1,ig))+bb(3,2)*(sk(2)+kg(2,ig)) &
               +bb(3,3)*(sk(3)+kg(3,ig))
          r = sqrt(x*x + y*y + z*z)
          x = x/r
          y = y/r
          z = z/r
          iypx = dcmplx(x,y)
          iymx = dcmplx(-x,y)
          ylpmj(1,ig,1) = 0
          ylpmj(2,ig,1) = 1
          ylpmj(1,ig,2) = 1
          ylpmj(2,ig,2) = 0
          ylpmj(1,ig,3) = iymx
          ylpmj(2,ig,3) = z
          ylpmj(1,ig,4) = -z
          ylpmj(2,ig,4) = -iypx
          ylpmj(1,ig,5) = 0
          ylpmj(2,ig,5) = -iymx*sq3/sq2
          ylpmj(1,ig,6) = -iymx/sq2
          ylpmj(2,ig,6) = sq2*z
          ylpmj(1,ig,7) = sq2*z
          ylpmj(2,ig,7) = -iypx/sq2
          ylpmj(1,ig,8) = -iypx*sq3/sq2
          ylpmj(2,ig,8) = 0
       end do
    case(3)
       do ig=ista,nkm
          x = bb(1,1)*(sk(1)+kg(1,ig))+bb(1,2)*(sk(2)+kg(2,ig)) &
               +bb(1,3)*(sk(3)+kg(3,ig))
          y = bb(2,1)*(sk(1)+kg(1,ig))+bb(2,2)*(sk(2)+kg(2,ig)) &
               +bb(2,3)*(sk(3)+kg(3,ig))
          z = bb(3,1)*(sk(1)+kg(1,ig))+bb(3,2)*(sk(2)+kg(2,ig)) &
               +bb(3,3)*(sk(3)+kg(3,ig))
          r = sqrt(x*x + y*y + z*z)
          x = x/r
          y = y/r
          z = z/r
          iypx = dcmplx(x,y)
          iymx = dcmplx(-x,y)
          ylpmj(1,ig,1) = 0
          ylpmj(2,ig,1) = 1
          ylpmj(1,ig,2) = 1
          ylpmj(2,ig,2) = 0
          ylpmj(1,ig,3) = iymx
          ylpmj(2,ig,3) = z
          ylpmj(1,ig,4) = -z
          ylpmj(2,ig,4) = -iypx
          ylpmj(1,ig,5) = 0
          ylpmj(2,ig,5) = -iymx*sq3/sq2
          ylpmj(1,ig,6) = -iymx/sq2
          ylpmj(2,ig,6) = sq2*z
          ylpmj(1,ig,7) = sq2*z
          ylpmj(2,ig,7) = -iypx/sq2
          ylpmj(1,ig,8) = -iypx*sq3/sq2
          ylpmj(2,ig,8) = 0
          ylpmj(1,ig,9) = -iymx**2*sq6/2.0d0
          ylpmj(2,ig,9) = -iymx*sq3*z/sq2
          ylpmj(1,ig,10) = 3*iymx*z/sq2
          ylpmj(2,ig,10) = (3*z**2-1)/sq2
          ylpmj(1,ig,11) = -(3*z**2-1)/sq2
          ylpmj(2,ig,11) = -3*iypx*z/sq2
          ylpmj(1,ig,12) = 3*iypx*z/sq6
          ylpmj(2,ig,12) = 3*iypx**2/sq6
          ylpmj(1,ig,13) = 0
          ylpmj(2,ig,13) = iymx**2*sq30/4.0d0
          ylpmj(1,ig,14) = iymx**2*sq6/4.0d0
          ylpmj(2,ig,14) = -iymx*sq6*z
          ylpmj(1,ig,15) = -iymx*sq3*z
          ylpmj(2,ig,15) = sq3*(3*z**2-1)/2.0d0
          ylpmj(1,ig,16) = sq3*(3*z**2-1)/2.0d0
          ylpmj(2,ig,16) = -iypx*sq3*z
          ylpmj(1,ig,17) = -iypx*sq6*z
          ylpmj(2,ig,17) = 3.0d0*iypx**2/(2.0d0*sq6)
          ylpmj(1,ig,18) = 3.0d0*iypx**2*sq5/(2.0d0*sq6)
          ylpmj(2,ig,18) = 0
       end do
    case(4)
       do ig=ista,nkm
          x = bb(1,1)*(sk(1)+kg(1,ig))+bb(1,2)*(sk(2)+kg(2,ig)) &
               +bb(1,3)*(sk(3)+kg(3,ig))
          y = bb(2,1)*(sk(1)+kg(1,ig))+bb(2,2)*(sk(2)+kg(2,ig)) &
               +bb(2,3)*(sk(3)+kg(3,ig))
          z = bb(3,1)*(sk(1)+kg(1,ig))+bb(3,2)*(sk(2)+kg(2,ig)) &
               +bb(3,3)*(sk(3)+kg(3,ig))
          r = sqrt(x*x + y*y + z*z)
          x = x/r
          y = y/r
          z = z/r
          iypx = dcmplx(x,y)
          iymx = dcmplx(-x,y)
          ylpmj(1,ig,1) = 0
          ylpmj(2,ig,1) = 1
          ylpmj(1,ig,2) = 1
          ylpmj(2,ig,2) = 0
          ylpmj(1,ig,3) = iymx
          ylpmj(2,ig,3) = z
          ylpmj(1,ig,4) = -z
          ylpmj(2,ig,4) = -iypx
          ylpmj(1,ig,5) = 0
          ylpmj(2,ig,5) = -iymx*sq3/sq2
          ylpmj(1,ig,6) = -iymx/sq2
          ylpmj(2,ig,6) = sq2*z
          ylpmj(1,ig,7) = sq2*z
          ylpmj(2,ig,7) = -iypx/sq2
          ylpmj(1,ig,8) = -iypx*sq3/sq2
          ylpmj(2,ig,8) = 0
          ylpmj(1,ig,9) = -iymx**2*sq6/2.0d0
          ylpmj(2,ig,9) = -iymx*sq3*z/sq2
          ylpmj(1,ig,10) = 3*iymx*z/sq2
          ylpmj(2,ig,10) = (3*z**2-1)/sq2
          ylpmj(1,ig,11) = -(3*z**2-1)/sq2
          ylpmj(2,ig,11) = -3*iypx*z/sq2
          ylpmj(1,ig,12) = 3*iypx*z/sq6
          ylpmj(2,ig,12) = 3*iypx**2/sq6
          ylpmj(1,ig,13) = 0
          ylpmj(2,ig,13) = iymx**2*sq30/4.0d0
          ylpmj(1,ig,14) = iymx**2*sq6/4.0d0
          ylpmj(2,ig,14) = -iymx*sq6*z
          ylpmj(1,ig,15) = -iymx*sq3*z
          ylpmj(2,ig,15) = sq3*(3*z**2-1)/2.0d0
          ylpmj(1,ig,16) = sq3*(3*z**2-1)/2.0d0
          ylpmj(2,ig,16) = -iypx*sq3*z
          ylpmj(1,ig,17) = -iypx*sq6*z
          ylpmj(2,ig,17) = 3.0d0*iypx**2/(2.0d0*sq6)
          ylpmj(1,ig,18) = 3.0d0*iypx**2*sq5/(2.0d0*sq6)
          ylpmj(2,ig,18) = 0
          ylpmj(1,ig,19) = iymx**3*sq30/4.0d0
          ylpmj(2,ig,19) = iymx**2*sq30*z/4.0d0
          ylpmj(1,ig,20) = (-5.0d0)*iymx**2*sq6*z/4.0d0
          ylpmj(2,ig,20) = -iymx*sq3*(5*z**2-1)/sq2/2.0d0
          ylpmj(1,ig,21) = iymx*sq3*(5*z**2-1)/2.0d0
          ylpmj(2,ig,21) = sq3*z*(5*z**2-3)/2.0d0
          ylpmj(1,ig,22) = -sq3*z*(5*z**2-3)/2.0d0
          ylpmj(2,ig,22) = -iypx*sq3*(5*z**2-1)/2.0d0
          ylpmj(1,ig,23) = iypx*sq3*(5*z**2-1)/sq2/2.0d0
          ylpmj(2,ig,23) = 15.0d0*iypx**2*z/(2.0d0*sq6)
          ylpmj(1,ig,24) = (-15.0d0)*iypx**2*z/(2.0d0*sq30)
          ylpmj(2,ig,24) = -iypx**3*sq30/4.0d0
          ylpmj(1,ig,25) = 0
          ylpmj(2,ig,25) = -iymx**3*sq35/4.0d0
          ylpmj(1,ig,26) = -iymx**3*sq5/4.0d0
          ylpmj(2,ig,26) = 3.0d0*iymx**2*sq5*z/2.0d0
          ylpmj(1,ig,27) = iymx**2*sq15*z/2.0d0
          ylpmj(2,ig,27) = -iymx*sq15*(5*z**2-1)/4.0d0
          ylpmj(1,ig,28) = (-3.0d0)*iymx*(5*z**2-1)/4.0d0
          ylpmj(2,ig,28) = z*(5*z**2-3)
          ylpmj(1,ig,29) = z*(5*z**2-3)
          ylpmj(2,ig,29) = (-3.0d0)*iypx*(5*z**2-1)/4.0d0
          ylpmj(1,ig,30) = -iypx*sq15*(5*z**2-1)/4.0d0
          ylpmj(2,ig,30) = iypx**2*sq15*z/2.0d0
          ylpmj(1,ig,31) = 3.0d0*iypx**2*sq5*z/2.0d0
          ylpmj(2,ig,31) = -iypx**3*sq5/4.0d0
          ylpmj(1,ig,32) = -iypx**3*sq35/4.0d0
          ylpmj(2,ig,32) = 0
       end do
    case default
       write(6,*) 'getylb_two: invalid lmax =', lmax
       stop
    end select

    ylpmj = (1.0d0/sqrt(4.0d0*pi))*ylpmj
    return
  end subroutine getylb_two

  subroutine getyls(nkm,sk,kg,bb,ylms,lmax,ncomp)
    implicit none
    integer,intent(in):: nkm,kg(3,nkm),lmax
    real(kind=8),intent(in):: bb(3,3),sk(3)
    complex(kind=8):: ylms(ncomp*nkm,3,ncomp*lmax**2)
    integer:: ncomp

    if (ncomp.eq.1) then
       call getyls_one(nkm,sk,kg,bb,ylms,lmax)
    else
       call getyls_two(nkm,sk,kg,bb,ylms,lmax)
    end if
    return
  end subroutine getyls

  subroutine getyls_one(nkm,sk,kg,bb,ylms,lmax)
    implicit none
    integer,intent(in):: nkm,kg(3,nkm),lmax
    real(kind=8),intent(in):: bb(3,3),sk(3)
    complex(kind=8):: ylms(nkm,3,lmax**2)

    real(kind=8),parameter:: pi = 3.14159265358979323846d0
    real(kind=8):: s1,s2,s3,s0,f10,f20,f21,f22,r
    real(kind=8):: f30,f31,f32,f33
    integer:: ista,ig

    s1=bb(1,1)*(sk(1)+kg(1,1))+bb(1,2)*(sk(2)+kg(2,1)) &
         +bb(1,3)*(sk(3)+kg(3,1))
    s2=bb(2,1)*(sk(1)+kg(1,1))+bb(2,2)*(sk(2)+kg(2,1)) &
         +bb(2,3)*(sk(3)+kg(3,1))
    s3=bb(3,1)*(sk(1)+kg(1,1))+bb(3,2)*(sk(2)+kg(2,1)) &
         +bb(3,3)*(sk(3)+kg(3,1))
    s0=s1*s1+s2*s2+s3*s3

    f10=sqrt( 3.d0 / ( 4.d0*pi) )
    f20=sqrt( 15.d0/pi )*0.5d0
    f21=sqrt( 5.0d0/pi )*0.25d0
    f22=f20*0.5d0
    !
    f30=sqrt(35.d0/(2.d0*pi))*0.25d0
    f31=sqrt(105.d0/pi)
    f32=sqrt(21.d0/(2.d0*pi))*0.25d0
    f33=sqrt(7.d0/pi)*0.25d0

    if(s0.eq.0.d0) then
       ylms(1,:,:) = 0.0d0
       ista=2
    else
       ista=1
    endif
    if (lmax.eq.1) then
       do ig=ista,nkm
          ylms(ig,:,1)= 0.0d0
       end do
    else if (lmax.eq.2) then
       do ig=ista,nkm
          s1=bb(1,1)*(sk(1)+kg(1,ig))+bb(1,2)*(sk(2)+kg(2,ig)) &
               +bb(1,3)*(sk(3)+kg(3,ig))
          s2=bb(2,1)*(sk(1)+kg(1,ig))+bb(2,2)*(sk(2)+kg(2,ig)) &
               +bb(2,3)*(sk(3)+kg(3,ig))
          s3=bb(3,1)*(sk(1)+kg(1,ig))+bb(3,2)*(sk(2)+kg(2,ig)) &
               +bb(3,3)*(sk(3)+kg(3,ig))
          s0=s1*s1+s2*s2+s3*s3
          r=sqrt(s0)
          ylms(ig,:,1)= 0.0d0
          ylms(ig,1,2)= 0.0d0
          ylms(ig,2,2)= -f10
          ylms(ig,3,2)= 0.0d0
          ylms(ig,1,3)= 0.0d0
          ylms(ig,2,3)= 0.0d0
          ylms(ig,3,3)= f10
          ylms(ig,1,4)= -f10
          ylms(ig,2,4)= 0.0d0
          ylms(ig,3,4)= 0.0d0
       end do
    else if (lmax.eq.3) then
       do ig=ista,nkm
          s1=bb(1,1)*(sk(1)+kg(1,ig))+bb(1,2)*(sk(2)+kg(2,ig)) &
               +bb(1,3)*(sk(3)+kg(3,ig))
          s2=bb(2,1)*(sk(1)+kg(1,ig))+bb(2,2)*(sk(2)+kg(2,ig)) &
               +bb(2,3)*(sk(3)+kg(3,ig))
          s3=bb(3,1)*(sk(1)+kg(1,ig))+bb(3,2)*(sk(2)+kg(2,ig)) &
               +bb(3,3)*(sk(3)+kg(3,ig))
          s0=s1*s1+s2*s2+s3*s3
          r=sqrt(s0)
          ylms(ig,:,1)= 0.0d0
          ylms(ig,1,2)= 0.0d0
          ylms(ig,2,2)= -f10
          ylms(ig,3,2)= 0.0d0
          ylms(ig,1,3)= 0.0d0
          ylms(ig,2,3)= 0.0d0
          ylms(ig,3,3)= f10
          ylms(ig,1,4)= -f10
          ylms(ig,2,4)= 0.0d0
          ylms(ig,3,4)= 0.0d0
          ylms(ig,1,5)= f20*s2/r
          ylms(ig,2,5)= f20*s1/r
          ylms(ig,3,5)= 0.0d0
          ylms(ig,1,6)= 0.0d0
          ylms(ig,2,6)= -f20*s3/r
          ylms(ig,3,6)= -f20*s2/r
          ylms(ig,1,7)= 0.0d0
          ylms(ig,2,7)= 0.0d0
          ylms(ig,3,7)= f21*6.0d0*s3/r
          ylms(ig,1,8)= -f20*s3/r
          ylms(ig,2,8)= 0.0d0
          ylms(ig,3,8)= -f20*s1/r
          ylms(ig,1,9)= f22*2.0d0*s1/r
          ylms(ig,2,9)= -f22*2.0d0*s2/r
          ylms(ig,3,9)= 0.0d0
       end do
    else if (lmax.eq.4) then
       do ig=ista,nkm
          s1=bb(1,1)*(sk(1)+kg(1,ig))+bb(1,2)*(sk(2)+kg(2,ig)) &
               +bb(1,3)*(sk(3)+kg(3,ig))
          s2=bb(2,1)*(sk(1)+kg(1,ig))+bb(2,2)*(sk(2)+kg(2,ig)) &
               +bb(2,3)*(sk(3)+kg(3,ig))
          s3=bb(3,1)*(sk(1)+kg(1,ig))+bb(3,2)*(sk(2)+kg(2,ig)) &
               +bb(3,3)*(sk(3)+kg(3,ig))
          s0=s1*s1+s2*s2+s3*s3
          r=sqrt(s0)
          s1=s1/r
          s2=s2/r
          s3=s3/r
          ylms(ig,:,1)= 0.0d0
          ylms(ig,1,2)= 0.0d0
          ylms(ig,2,2)= -f10
          ylms(ig,3,2)= 0.0d0
          ylms(ig,1,3)= 0.0d0
          ylms(ig,2,3)= 0.0d0
          ylms(ig,3,3)= f10
          ylms(ig,1,4)= -f10
          ylms(ig,2,4)= 0.0d0
          ylms(ig,3,4)= 0.0d0
          ylms(ig,1,5)= f20*s2
          ylms(ig,2,5)= f20*s1
          ylms(ig,3,5)= 0.0d0
          ylms(ig,1,6)= 0.0d0
          ylms(ig,2,6)= -f20*s3
          ylms(ig,3,6)= -f20*s2
          ylms(ig,1,7)= 0.0d0
          ylms(ig,2,7)= 0.0d0
          ylms(ig,3,7)= f21*6.0d0*s3
          ylms(ig,1,8)= -f20*s3
          ylms(ig,2,8)= 0.0d0
          ylms(ig,3,8)= -f20*s1
          ylms(ig,1,9)= f22*2.0d0*s1
          ylms(ig,2,9)= -f22*2.0d0*s2
          ylms(ig,3,9)= 0.0d0
          ylms(ig,1,10)= -6*f30*s1*s2
          ylms(ig,2,10)= 2*f30*s2**2-f30*(3*s1**2-s2**2)
          ylms(ig,3,10)= 0
          ylms(ig,1,11)= f31*s2*s3/2.0d0
          ylms(ig,2,11)= f31*s1*s3/2.0d0
          ylms(ig,3,11)= f31*s1*s2/2.0d0
          ylms(ig,1,12)= 0
          ylms(ig,2,12)= -f32*(5*s3**2-1)
          ylms(ig,3,12)= -10*f32*s2*s3
          ylms(ig,1,13)= 0
          ylms(ig,2,13)= 0
          ylms(ig,3,13)= f33*(5*s3**2-3)+10*f33*s3**2
          ylms(ig,1,14)= -f32*(5*s3**2-1)
          ylms(ig,2,14)= 0
          ylms(ig,3,14)= -10*f32*s1*s3
          ylms(ig,1,15)= f31*s1*s3/2.0d0
          ylms(ig,2,15)= -f31*s2*s3/2.0d0
          ylms(ig,3,15)= f31*(s1**2-s2**2)/4.0d0
          ylms(ig,1,16)= -f30*(s1**2-3*s2**2)-2*f30*s1**2
          ylms(ig,2,16)= 6*f30*s1*s2
          ylms(ig,3,16)= 0
       end do
    else
       write(6,*)'LMAX must be <= 4.  LMAX=',lmax
       stop 'GETYLS'
    end if
    return
  end subroutine getyls_one

  subroutine getyls_two(nkm,sk,kg,bb,ylpmjs,lmax)
    implicit none
    integer,intent(in):: nkm,kg(3,nkm),lmax
    real(kind=8),intent(in):: bb(3,3),sk(3)
    complex(kind=8):: ylpmjs(2,nkm,3,2*lmax**2)

    real(kind=8),parameter:: pi = 3.14159265358979323846d0
    complex(kind=8),parameter:: ci = (0.0d0, 1.0d0)

    real(kind=8):: x,y,z,r
    integer:: ista,ig
    complex(kind=8):: iypx, iymx

    real(kind=8),parameter:: sq2 = sqrt(2.0d0), &
         sq3 = sqrt(3.0d0), sq5 = sqrt(5.0d0), &
         sq6 = sqrt(6.0d0), sq8 = sqrt(8.0d0), &
         sq15 = sqrt(15.0d0), sq30 = sqrt(30.0d0), &
         sq35 = sqrt(35.0d0)

    x = bb(1,1)*(sk(1)+kg(1,1))+bb(1,2)*(sk(2)+kg(2,1)) &
         +bb(1,3)*(sk(3)+kg(3,1))
    y = bb(2,1)*(sk(1)+kg(1,1))+bb(2,2)*(sk(2)+kg(2,1)) &
         +bb(2,3)*(sk(3)+kg(3,1))
    z = bb(3,1)*(sk(1)+kg(1,1))+bb(3,2)*(sk(2)+kg(2,1)) &
         +bb(3,3)*(sk(3)+kg(3,1))
    r = x*x + y*y + z*z

    if(r.eq.0.d0) then
       ylpmjs(:,1,:,:) = 0.0d0
       ista=2
    else
       ista=1
    endif

    select case(lmax)
    case(1)
       do ig=ista,nkm
          ylpmjs(1,ig,1,1) = 0
          ylpmjs(2,ig,1,1) = 0
          ylpmjs(1,ig,2,1) = 0
          ylpmjs(2,ig,2,1) = 0
          ylpmjs(1,ig,3,1) = 0
          ylpmjs(2,ig,3,1) = 0
          ylpmjs(1,ig,1,2) = 0
          ylpmjs(2,ig,1,2) = 0
          ylpmjs(1,ig,2,2) = 0
          ylpmjs(2,ig,2,2) = 0
          ylpmjs(1,ig,3,2) = 0
          ylpmjs(2,ig,3,2) = 0
       end do
    case(2)
       do ig=ista,nkm
          x = bb(1,1)*(sk(1)+kg(1,ig))+bb(1,2)*(sk(2)+kg(2,ig)) &
               +bb(1,3)*(sk(3)+kg(3,ig))
          y = bb(2,1)*(sk(1)+kg(1,ig))+bb(2,2)*(sk(2)+kg(2,ig)) &
               +bb(2,3)*(sk(3)+kg(3,ig))
          z = bb(3,1)*(sk(1)+kg(1,ig))+bb(3,2)*(sk(2)+kg(2,ig)) &
               +bb(3,3)*(sk(3)+kg(3,ig))
          r = sqrt(x*x + y*y + z*z)
          x = x/r
          y = y/r
          z = z/r
          iypx = dcmplx(x,y)
          iymx = dcmplx(-x,y)
          ylpmjs(1,ig,1,1) = 0
          ylpmjs(2,ig,1,1) = 0
          ylpmjs(1,ig,2,1) = 0
          ylpmjs(2,ig,2,1) = 0
          ylpmjs(1,ig,3,1) = 0
          ylpmjs(2,ig,3,1) = 0
          ylpmjs(1,ig,1,2) = 0
          ylpmjs(2,ig,1,2) = 0
          ylpmjs(1,ig,2,2) = 0
          ylpmjs(2,ig,2,2) = 0
          ylpmjs(1,ig,3,2) = 0
          ylpmjs(2,ig,3,2) = 0
          ylpmjs(1,ig,1,3) = -1
          ylpmjs(2,ig,1,3) = 0
          ylpmjs(1,ig,2,3) = ci
          ylpmjs(2,ig,2,3) = 0
          ylpmjs(1,ig,3,3) = 0
          ylpmjs(2,ig,3,3) = 1
          ylpmjs(1,ig,1,4) = 0
          ylpmjs(2,ig,1,4) = -1
          ylpmjs(1,ig,2,4) = 0
          ylpmjs(2,ig,2,4) = -ci
          ylpmjs(1,ig,3,4) = -1
          ylpmjs(2,ig,3,4) = 0
          ylpmjs(1,ig,1,5) = 0
          ylpmjs(2,ig,1,5) = sq3/sq2
          ylpmjs(1,ig,2,5) = 0
          ylpmjs(2,ig,2,5) = -ci*sq3/sq2
          ylpmjs(1,ig,3,5) = 0
          ylpmjs(2,ig,3,5) = 0
          ylpmjs(1,ig,1,6) = 1/sq2
          ylpmjs(2,ig,1,6) = 0
          ylpmjs(1,ig,2,6) = -ci/sq2
          ylpmjs(2,ig,2,6) = 0
          ylpmjs(1,ig,3,6) = 0
          ylpmjs(2,ig,3,6) = sq2
          ylpmjs(1,ig,1,7) = 0
          ylpmjs(2,ig,1,7) = -1/sq2
          ylpmjs(1,ig,2,7) = 0
          ylpmjs(2,ig,2,7) = -ci/sq2
          ylpmjs(1,ig,3,7) = sq2
          ylpmjs(2,ig,3,7) = 0
          ylpmjs(1,ig,1,8) = -sq3/sq2
          ylpmjs(2,ig,1,8) = 0
          ylpmjs(1,ig,2,8) = -ci*sq3/sq2
          ylpmjs(2,ig,2,8) = 0
          ylpmjs(1,ig,3,8) = 0
          ylpmjs(2,ig,3,8) = 0
       end do
    case(3)
       do ig=ista,nkm
          x = bb(1,1)*(sk(1)+kg(1,ig))+bb(1,2)*(sk(2)+kg(2,ig)) &
               +bb(1,3)*(sk(3)+kg(3,ig))
          y = bb(2,1)*(sk(1)+kg(1,ig))+bb(2,2)*(sk(2)+kg(2,ig)) &
               +bb(2,3)*(sk(3)+kg(3,ig))
          z = bb(3,1)*(sk(1)+kg(1,ig))+bb(3,2)*(sk(2)+kg(2,ig)) &
               +bb(3,3)*(sk(3)+kg(3,ig))
          r = sqrt(x*x + y*y + z*z)
          x = x/r
          y = y/r
          z = z/r
          iypx = dcmplx(x,y)
          iymx = dcmplx(-x,y)
          ylpmjs(1,ig,1,1) = 0
          ylpmjs(2,ig,1,1) = 0
          ylpmjs(1,ig,2,1) = 0
          ylpmjs(2,ig,2,1) = 0
          ylpmjs(1,ig,3,1) = 0
          ylpmjs(2,ig,3,1) = 0
          ylpmjs(1,ig,1,2) = 0
          ylpmjs(2,ig,1,2) = 0
          ylpmjs(1,ig,2,2) = 0
          ylpmjs(2,ig,2,2) = 0
          ylpmjs(1,ig,3,2) = 0
          ylpmjs(2,ig,3,2) = 0
          ylpmjs(1,ig,1,3) = -1
          ylpmjs(2,ig,1,3) = 0
          ylpmjs(1,ig,2,3) = ci
          ylpmjs(2,ig,2,3) = 0
          ylpmjs(1,ig,3,3) = 0
          ylpmjs(2,ig,3,3) = 1
          ylpmjs(1,ig,1,4) = 0
          ylpmjs(2,ig,1,4) = -1
          ylpmjs(1,ig,2,4) = 0
          ylpmjs(2,ig,2,4) = -ci
          ylpmjs(1,ig,3,4) = -1
          ylpmjs(2,ig,3,4) = 0
          ylpmjs(1,ig,1,5) = 0
          ylpmjs(2,ig,1,5) = sq3/sq2
          ylpmjs(1,ig,2,5) = 0
          ylpmjs(2,ig,2,5) = -ci*sq3/sq2
          ylpmjs(1,ig,3,5) = 0
          ylpmjs(2,ig,3,5) = 0
          ylpmjs(1,ig,1,6) = 1/sq2
          ylpmjs(2,ig,1,6) = 0
          ylpmjs(1,ig,2,6) = -ci/sq2
          ylpmjs(2,ig,2,6) = 0
          ylpmjs(1,ig,3,6) = 0
          ylpmjs(2,ig,3,6) = sq2
          ylpmjs(1,ig,1,7) = 0
          ylpmjs(2,ig,1,7) = -1/sq2
          ylpmjs(1,ig,2,7) = 0
          ylpmjs(2,ig,2,7) = -ci/sq2
          ylpmjs(1,ig,3,7) = sq2
          ylpmjs(2,ig,3,7) = 0
          ylpmjs(1,ig,1,8) = -sq3/sq2
          ylpmjs(2,ig,1,8) = 0
          ylpmjs(1,ig,2,8) = -ci*sq3/sq2
          ylpmjs(2,ig,2,8) = 0
          ylpmjs(1,ig,3,8) = 0
          ylpmjs(2,ig,3,8) = 0
          ylpmjs(1,ig,1,9) = iymx*sq6
          ylpmjs(2,ig,1,9) = sq3*z/sq2
          ylpmjs(1,ig,2,9) = -ci*iymx*sq6
          ylpmjs(2,ig,2,9) = -ci*sq3*z/sq2
          ylpmjs(1,ig,3,9) = 0
          ylpmjs(2,ig,3,9) = -iymx*sq3/sq2
          ylpmjs(1,ig,1,10) = -3*z/sq2
          ylpmjs(2,ig,1,10) = 0
          ylpmjs(1,ig,2,10) = 3*ci*z/sq2
          ylpmjs(2,ig,2,10) = 0
          ylpmjs(1,ig,3,10) = 3*iymx/sq2
          ylpmjs(2,ig,3,10) = 6*z/sq2
          ylpmjs(1,ig,1,11) = 0
          ylpmjs(2,ig,1,11) = -3*z/sq2
          ylpmjs(1,ig,2,11) = 0
          ylpmjs(2,ig,2,11) = -3*ci*z/sq2
          ylpmjs(1,ig,3,11) = -6*z/sq2
          ylpmjs(2,ig,3,11) = -3*iypx/sq2
          ylpmjs(1,ig,1,12) = 3*z/sq6
          ylpmjs(2,ig,1,12) = 6*iypx/sq6
          ylpmjs(1,ig,2,12) = 3*ci*z/sq6
          ylpmjs(2,ig,2,12) = 6*ci*iypx/sq6
          ylpmjs(1,ig,3,12) = 3*iypx/sq6
          ylpmjs(2,ig,3,12) = 0
          ylpmjs(1,ig,1,13) = 0
          ylpmjs(2,ig,1,13) = -iymx*sq30/2.0d0
          ylpmjs(1,ig,2,13) = 0
          ylpmjs(2,ig,2,13) = ci*iymx*sq30/2.0d0
          ylpmjs(1,ig,3,13) = 0
          ylpmjs(2,ig,3,13) = 0
          ylpmjs(1,ig,1,14) = -iymx*sq6/2.0d0
          ylpmjs(2,ig,1,14) = sq6*z
          ylpmjs(1,ig,2,14) = ci*iymx*sq6/2.0d0
          ylpmjs(2,ig,2,14) = -ci*sq6*z
          ylpmjs(1,ig,3,14) = 0
          ylpmjs(2,ig,3,14) = -iymx*sq6
          ylpmjs(1,ig,1,15) = sq3*z
          ylpmjs(2,ig,1,15) = 0
          ylpmjs(1,ig,2,15) = -ci*sq3*z
          ylpmjs(2,ig,2,15) = 0
          ylpmjs(1,ig,3,15) = -iymx*sq3
          ylpmjs(2,ig,3,15) = 3*sq3*z
          ylpmjs(1,ig,1,16) = 0
          ylpmjs(2,ig,1,16) = -sq3*z
          ylpmjs(1,ig,2,16) = 0
          ylpmjs(2,ig,2,16) = -ci*sq3*z
          ylpmjs(1,ig,3,16) = 3*sq3*z
          ylpmjs(2,ig,3,16) = -iypx*sq3
          ylpmjs(1,ig,1,17) = -sq6*z
          ylpmjs(2,ig,1,17) = 3*iypx/sq6
          ylpmjs(1,ig,2,17) = -ci*sq6*z
          ylpmjs(2,ig,2,17) = 3*ci*iypx/sq6
          ylpmjs(1,ig,3,17) = -iypx*sq6
          ylpmjs(2,ig,3,17) = 0
          ylpmjs(1,ig,1,18) = 3*iypx*sq5/sq6
          ylpmjs(2,ig,1,18) = 0
          ylpmjs(1,ig,2,18) = 3*ci*iypx*sq5/sq6
          ylpmjs(2,ig,2,18) = 0
          ylpmjs(1,ig,3,18) = 0
          ylpmjs(2,ig,3,18) = 0
       end do
    case(4)
       do ig=ista,nkm
          x = bb(1,1)*(sk(1)+kg(1,ig))+bb(1,2)*(sk(2)+kg(2,ig)) &
               +bb(1,3)*(sk(3)+kg(3,ig))
          y = bb(2,1)*(sk(1)+kg(1,ig))+bb(2,2)*(sk(2)+kg(2,ig)) &
               +bb(2,3)*(sk(3)+kg(3,ig))
          z = bb(3,1)*(sk(1)+kg(1,ig))+bb(3,2)*(sk(2)+kg(2,ig)) &
               +bb(3,3)*(sk(3)+kg(3,ig))
          r = sqrt(x*x + y*y + z*z)
          x = x/r
          y = y/r
          z = z/r
          iypx = dcmplx(x,y)
          iymx = dcmplx(-x,y)
          ylpmjs(1,ig,1,1) = 0
          ylpmjs(2,ig,1,1) = 0
          ylpmjs(1,ig,2,1) = 0
          ylpmjs(2,ig,2,1) = 0
          ylpmjs(1,ig,3,1) = 0
          ylpmjs(2,ig,3,1) = 0
          ylpmjs(1,ig,1,2) = 0
          ylpmjs(2,ig,1,2) = 0
          ylpmjs(1,ig,2,2) = 0
          ylpmjs(2,ig,2,2) = 0
          ylpmjs(1,ig,3,2) = 0
          ylpmjs(2,ig,3,2) = 0
          ylpmjs(1,ig,1,3) = -1
          ylpmjs(2,ig,1,3) = 0
          ylpmjs(1,ig,2,3) = ci
          ylpmjs(2,ig,2,3) = 0
          ylpmjs(1,ig,3,3) = 0
          ylpmjs(2,ig,3,3) = 1
          ylpmjs(1,ig,1,4) = 0
          ylpmjs(2,ig,1,4) = -1
          ylpmjs(1,ig,2,4) = 0
          ylpmjs(2,ig,2,4) = -ci
          ylpmjs(1,ig,3,4) = -1
          ylpmjs(2,ig,3,4) = 0
          ylpmjs(1,ig,1,5) = 0
          ylpmjs(2,ig,1,5) = sq3/sq2
          ylpmjs(1,ig,2,5) = 0
          ylpmjs(2,ig,2,5) = -ci*sq3/sq2
          ylpmjs(1,ig,3,5) = 0
          ylpmjs(2,ig,3,5) = 0
          ylpmjs(1,ig,1,6) = 1/sq2
          ylpmjs(2,ig,1,6) = 0
          ylpmjs(1,ig,2,6) = -ci/sq2
          ylpmjs(2,ig,2,6) = 0
          ylpmjs(1,ig,3,6) = 0
          ylpmjs(2,ig,3,6) = sq2
          ylpmjs(1,ig,1,7) = 0
          ylpmjs(2,ig,1,7) = -1/sq2
          ylpmjs(1,ig,2,7) = 0
          ylpmjs(2,ig,2,7) = -ci/sq2
          ylpmjs(1,ig,3,7) = sq2
          ylpmjs(2,ig,3,7) = 0
          ylpmjs(1,ig,1,8) = -sq3/sq2
          ylpmjs(2,ig,1,8) = 0
          ylpmjs(1,ig,2,8) = -ci*sq3/sq2
          ylpmjs(2,ig,2,8) = 0
          ylpmjs(1,ig,3,8) = 0
          ylpmjs(2,ig,3,8) = 0
          ylpmjs(1,ig,1,9) = iymx*sq6
          ylpmjs(2,ig,1,9) = sq3*z/sq2
          ylpmjs(1,ig,2,9) = -ci*iymx*sq6
          ylpmjs(2,ig,2,9) = -ci*sq3*z/sq2
          ylpmjs(1,ig,3,9) = 0
          ylpmjs(2,ig,3,9) = -iymx*sq3/sq2
          ylpmjs(1,ig,1,10) = -3*z/sq2
          ylpmjs(2,ig,1,10) = 0
          ylpmjs(1,ig,2,10) = 3*ci*z/sq2
          ylpmjs(2,ig,2,10) = 0
          ylpmjs(1,ig,3,10) = 3*iymx/sq2
          ylpmjs(2,ig,3,10) = 6*z/sq2
          ylpmjs(1,ig,1,11) = 0
          ylpmjs(2,ig,1,11) = -3*z/sq2
          ylpmjs(1,ig,2,11) = 0
          ylpmjs(2,ig,2,11) = -3*ci*z/sq2
          ylpmjs(1,ig,3,11) = -6*z/sq2
          ylpmjs(2,ig,3,11) = -3*iypx/sq2
          ylpmjs(1,ig,1,12) = 3*z/sq6
          ylpmjs(2,ig,1,12) = 6*iypx/sq6
          ylpmjs(1,ig,2,12) = 3*ci*z/sq6
          ylpmjs(2,ig,2,12) = 6*ci*iypx/sq6
          ylpmjs(1,ig,3,12) = 3*iypx/sq6
          ylpmjs(2,ig,3,12) = 0
          ylpmjs(1,ig,1,13) = 0
          ylpmjs(2,ig,1,13) = -iymx*sq30/2.0d0
          ylpmjs(1,ig,2,13) = 0
          ylpmjs(2,ig,2,13) = ci*iymx*sq30/2.0d0
          ylpmjs(1,ig,3,13) = 0
          ylpmjs(2,ig,3,13) = 0
          ylpmjs(1,ig,1,14) = -iymx*sq6/2.0d0
          ylpmjs(2,ig,1,14) = sq6*z
          ylpmjs(1,ig,2,14) = ci*iymx*sq6/2.0d0
          ylpmjs(2,ig,2,14) = -ci*sq6*z
          ylpmjs(1,ig,3,14) = 0
          ylpmjs(2,ig,3,14) = -iymx*sq6
          ylpmjs(1,ig,1,15) = sq3*z
          ylpmjs(2,ig,1,15) = 0
          ylpmjs(1,ig,2,15) = -ci*sq3*z
          ylpmjs(2,ig,2,15) = 0
          ylpmjs(1,ig,3,15) = -iymx*sq3
          ylpmjs(2,ig,3,15) = 3*sq3*z
          ylpmjs(1,ig,1,16) = 0
          ylpmjs(2,ig,1,16) = -sq3*z
          ylpmjs(1,ig,2,16) = 0
          ylpmjs(2,ig,2,16) = -ci*sq3*z
          ylpmjs(1,ig,3,16) = 3*sq3*z
          ylpmjs(2,ig,3,16) = -iypx*sq3
          ylpmjs(1,ig,1,17) = -sq6*z
          ylpmjs(2,ig,1,17) = 3*iypx/sq6
          ylpmjs(1,ig,2,17) = -ci*sq6*z
          ylpmjs(2,ig,2,17) = 3*ci*iypx/sq6
          ylpmjs(1,ig,3,17) = -iypx*sq6
          ylpmjs(2,ig,3,17) = 0
          ylpmjs(1,ig,1,18) = 3*iypx*sq5/sq6
          ylpmjs(2,ig,1,18) = 0
          ylpmjs(1,ig,2,18) = 3*ci*iypx*sq5/sq6
          ylpmjs(2,ig,2,18) = 0
          ylpmjs(1,ig,3,18) = 0
          ylpmjs(2,ig,3,18) = 0
          ylpmjs(1,ig,1,19) = (-3.0d0)*iymx**2*sq30/4.0d0
          ylpmjs(2,ig,1,19) = -iymx*sq30*z/2.0d0
          ylpmjs(1,ig,2,19) = 3.0d0*ci*iymx**2*sq30/4.0d0
          ylpmjs(2,ig,2,19) = ci*iymx*sq30*z/2.0d0
          ylpmjs(1,ig,3,19) = 0
          ylpmjs(2,ig,3,19) = iymx**2*sq30/4.0d0
          ylpmjs(1,ig,1,20) = 5.0d0*iymx*sq6*z/2.0d0
          ylpmjs(2,ig,1,20) = sq3*(5*z**2-1)/sq2/2.0d0
          ylpmjs(1,ig,2,20) = (-5.0d0)*ci*iymx*sq6*z/2.0d0
          ylpmjs(2,ig,2,20) = -ci*sq3*(5*z**2-1)/sq2/2.0d0
          ylpmjs(1,ig,3,20) = (-5.0d0)*iymx**2*sq6/4.0d0
          ylpmjs(2,ig,3,20) = -5*iymx*sq3*z/sq2
          ylpmjs(1,ig,1,21) = -sq3*(5*z**2-1)/2.0d0
          ylpmjs(2,ig,1,21) = 0
          ylpmjs(1,ig,2,21) = ci*sq3*(5*z**2-1)/2.0d0
          ylpmjs(2,ig,2,21) = 0
          ylpmjs(1,ig,3,21) = 5*iymx*sq3*z
          ylpmjs(2,ig,3,21) = 3.0d0*sq3*(5*z**2-1)/2.0d0
          ylpmjs(1,ig,1,22) = 0
          ylpmjs(2,ig,1,22) = -sq3*(5*z**2-1)/2.0d0
          ylpmjs(1,ig,2,22) = 0
          ylpmjs(2,ig,2,22) = -ci*sq3*(5*z**2-1)/2.0d0
          ylpmjs(1,ig,3,22) = (-3.0d0)*sq3*(5*z**2-1)/2.0d0
          ylpmjs(2,ig,3,22) = -5*iypx*sq3*z
          ylpmjs(1,ig,1,23) = sq3*(5*z**2-1)/sq2/2.0d0
          ylpmjs(2,ig,1,23) = 15*iypx*z/sq6
          ylpmjs(1,ig,2,23) = ci*sq3*(5*z**2-1)/sq2/2.0d0
          ylpmjs(2,ig,2,23) = 15*ci*iypx*z/sq6
          ylpmjs(1,ig,3,23) = 5*iypx*sq3*z/sq2
          ylpmjs(2,ig,3,23) = 15.0d0*iypx**2/(2.0d0*sq6)
          ylpmjs(1,ig,1,24) = -15*iypx*z/sq30
          ylpmjs(2,ig,1,24) = (-3.0d0)*iypx**2*sq30/4.0d0
          ylpmjs(1,ig,2,24) = -15*ci*iypx*z/sq30
          ylpmjs(2,ig,2,24) = (-3.0d0)*ci*iypx**2*sq30/4.0d0
          ylpmjs(1,ig,3,24) = (-15.0d0)*iypx**2/(2.0d0*sq30)
          ylpmjs(2,ig,3,24) = 0
          ylpmjs(1,ig,1,25) = 0
          ylpmjs(2,ig,1,25) = 3.0d0*iymx**2*sq35/4.0d0
          ylpmjs(1,ig,2,25) = 0
          ylpmjs(2,ig,2,25) = (-3.0d0)*ci*iymx**2*sq35/4.0d0
          ylpmjs(1,ig,3,25) = 0
          ylpmjs(2,ig,3,25) = 0
          ylpmjs(1,ig,1,26) = 3.0d0*iymx**2*sq5/4.0d0
          ylpmjs(2,ig,1,26) = -3*iymx*sq5*z
          ylpmjs(1,ig,2,26) = (-3.0d0)*ci*iymx**2*sq5/4.0d0
          ylpmjs(2,ig,2,26) = 3*ci*iymx*sq5*z
          ylpmjs(1,ig,3,26) = 0
          ylpmjs(2,ig,3,26) = 3.0d0*iymx**2*sq5/2.0d0
          ylpmjs(1,ig,1,27) = -iymx*sq15*z
          ylpmjs(2,ig,1,27) = sq15*(5*z**2-1)/4.0d0
          ylpmjs(1,ig,2,27) = ci*iymx*sq15*z
          ylpmjs(2,ig,2,27) = -ci*sq15*(5*z**2-1)/4.0d0
          ylpmjs(1,ig,3,27) = iymx**2*sq15/2.0d0
          ylpmjs(2,ig,3,27) = (-5.0d0)*iymx*sq15*z/2.0d0
          ylpmjs(1,ig,1,28) = 3.0d0*(5*z**2-1)/4.0d0
          ylpmjs(2,ig,1,28) = 0
          ylpmjs(1,ig,2,28) = (-3.0d0)*ci*(5*z**2-1)/4.0d0
          ylpmjs(2,ig,2,28) = 0
          ylpmjs(1,ig,3,28) = (-15.0d0)*iymx*z/2.0d0
          ylpmjs(2,ig,3,28) = 3*(5*z**2-1)
          ylpmjs(1,ig,1,29) = 0
          ylpmjs(2,ig,1,29) = (-3.0d0)*(5*z**2-1)/4.0d0
          ylpmjs(1,ig,2,29) = 0
          ylpmjs(2,ig,2,29) = (-3.0d0)*ci*(5*z**2-1)/4.0d0
          ylpmjs(1,ig,3,29) = 3*(5*z**2-1)
          ylpmjs(2,ig,3,29) = (-15.0d0)*iypx*z/2.0d0
          ylpmjs(1,ig,1,30) = -sq15*(5*z**2-1)/4.0d0
          ylpmjs(2,ig,1,30) = iypx*sq15*z
          ylpmjs(1,ig,2,30) = -ci*sq15*(5*z**2-1)/4.0d0
          ylpmjs(2,ig,2,30) = ci*iypx*sq15*z
          ylpmjs(1,ig,3,30) = (-5.0d0)*iypx*sq15*z/2.0d0
          ylpmjs(2,ig,3,30) = iypx**2*sq15/2.0d0
          ylpmjs(1,ig,1,31) = 3*iypx*sq5*z
          ylpmjs(2,ig,1,31) = (-3.0d0)*iypx**2*sq5/4.0d0
          ylpmjs(1,ig,2,31) = 3*ci*iypx*sq5*z
          ylpmjs(2,ig,2,31) = (-3.0d0)*ci*iypx**2*sq5/4.0d0
          ylpmjs(1,ig,3,31) = 3.0d0*iypx**2*sq5/2.0d0
          ylpmjs(2,ig,3,31) = 0
          ylpmjs(1,ig,1,32) = (-3.0d0)*iypx**2*sq35/4.0d0
          ylpmjs(2,ig,1,32) = 0
          ylpmjs(1,ig,2,32) = (-3.0d0)*ci*iypx**2*sq35/4.0d0
          ylpmjs(2,ig,2,32) = 0
          ylpmjs(1,ig,3,32) = 0
          ylpmjs(2,ig,3,32) = 0
       end do
    case default
       write(6,*) 'getyls_two: invalid lmax =', lmax
       stop
    end select

    ylpmjs = (1.0d0/sqrt(4.0d0*pi))*ylpmjs
    return
  end subroutine getyls_two

  subroutine formod(for,dfor,bb,nsi)
    implicit none
    integer nsi
    double precision for(3,nsi),dfor(3,nsi),bb(3,3)
    double precision cc(3,3)
    double precision pai, dfpai
    parameter(pai = 3.14159265358979324d0)
    parameter(dfpai=1.0d0/(4.0d0*pai*pai))

    integer i,j,k,it

    do i=1,3
       do j=1,3
          cc(i,j)=0.0d0
          do k=1,3
             cc(i,j)=cc(i,j)+bb(k,i)*bb(k,j)*dfpai
          end do
       end do
    end do
    do it=1,nsi
       do i=1,3
          dfor(i,it)=cc(i,1)*for(1,it)+cc(i,2)*for(2,it)+cc(i,3)*for(3,it)
       end do
    end do
    return
  end subroutine formod

  !  VTRAN
  !   transform vector by TT(3,3)
  !     V(3,*) <-   TT(3,3)*V(3,*)
  !
  subroutine vtran(v,tt,n)
    implicit none
    integer:: n
    real(kind=8):: v(3,n),tt(3,3),a(3)
    integer:: i,j

    do i=1,n
       do j=1,3
          a(j)=tt(j,1)*v(1,i)+tt(j,2)*v(2,i)+tt(j,3)*v(3,i)
       end do
       do j=1,3
          v(j,i)=a(j)
       end do
    end do
    return
  end subroutine vtran

  subroutine cell2cart(aa,n,vcell,vcart)
    implicit none
    integer:: n
    real(kind=8):: aa(3,3),vcell(3,n),vcart(3,n)
    integer:: i,j,k

    vcart = 0.0d0
    do i=1,n
       do j=1,3
          do k=1,3
             vcart(k,i) = vcart(k,i) + aa(k,j)*vcell(j,i)
          end do
       end do
    end do
  end subroutine cell2cart

  subroutine cart2cell(aa,n,vcart,vcell)
    implicit none
    integer:: n
    real(kind=8):: aa(3,3),vcell(3,n),vcart(3,n)
    integer:: i,j,k
    real(kind=8):: aai(3,3),w

    w=aa(1,1)*(aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3)) &
         +aa(2,1)*(aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3)) &
         +aa(3,1)*(aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3))

    aai(1,1)=(aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3))/w
    aai(1,2)=(aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3))/w
    aai(1,3)=(aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3))/w
    aai(2,1)=(aa(2,3)*aa(3,1)-aa(3,3)*aa(2,1))/w
    aai(2,2)=(aa(3,3)*aa(1,1)-aa(1,3)*aa(3,1))/w
    aai(2,3)=(aa(1,3)*aa(2,1)-aa(2,3)*aa(1,1))/w
    aai(3,1)=(aa(2,1)*aa(3,2)-aa(3,1)*aa(2,2))/w
    aai(3,2)=(aa(3,1)*aa(1,2)-aa(1,1)*aa(3,2))/w
    aai(3,3)=(aa(1,1)*aa(2,2)-aa(2,1)*aa(1,2))/w

    vcell = 0.0d0
    do i=1,n
       do j=1,3
          do k=1,3
             vcell(k,i) = vcell(k,i) + aai(k,j)*vcart(j,i)
          end do
       end do
    end do
  end subroutine cart2cell

  subroutine cart2cellr(aa,n,vcart,vcellr)
    implicit none
    integer:: n
    real(kind=8):: aa(3,3),vcellr(3,n),vcart(3,n)
    integer:: i,j,k
    real(kind=8):: aai(3,3),w

    w=aa(1,1)*(aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3)) &
         +aa(2,1)*(aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3)) &
         +aa(3,1)*(aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3))

    aai(1,1)=(aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3))/w
    aai(1,2)=(aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3))/w
    aai(1,3)=(aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3))/w
    aai(2,1)=(aa(2,3)*aa(3,1)-aa(3,3)*aa(2,1))/w
    aai(2,2)=(aa(3,3)*aa(1,1)-aa(1,3)*aa(3,1))/w
    aai(2,3)=(aa(1,3)*aa(2,1)-aa(2,3)*aa(1,1))/w
    aai(3,1)=(aa(2,1)*aa(3,2)-aa(3,1)*aa(2,2))/w
    aai(3,2)=(aa(3,1)*aa(1,2)-aa(1,1)*aa(3,2))/w
    aai(3,3)=(aa(1,1)*aa(2,2)-aa(2,1)*aa(1,2))/w

    vcellr = 0.0d0
    do i=1,n
       do j=1,3
          do k=1,3
             vcellr(k,i) = vcellr(k,i) + aai(k,j)*vcart(j,i)
          end do
       end do
       do k=1,3
          vcellr(k,i) = modulo(vcellr(k,i),1.0d0)
       end do
    end do
  end subroutine cart2cellr

  !
  ! transform contravariant vectors
  ! from cartesian coordinate to cell coordinate
  !
  subroutine acart2cell(aa,n,vcart,vcell)
    implicit none
    integer:: n
    real(kind=8):: aa(3,3),vcart(3,n),vcell(3,n)
    integer:: i,j,k

    vcell = 0.0d0
    do i=1,n
       do j=1,3
          do k=1,3
             vcell(k,i) = vcell(k,i) + aa(j,k)*vcart(j,i)
          end do
       end do
    end do
  end subroutine acart2cell

  subroutine recp_uvec(aa,bb)
    implicit none
    real(kind=8),intent(in):: aa(3,3)
    real(kind=8),intent(out):: bb(3,3)
    real(kind=8),parameter:: pi = 3.1415926535897932385d0

    real(kind=8):: v

    v=aa(1,1)*(aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3)) &
         +aa(2,1)*(aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3)) &
         +aa(3,1)*(aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3))
    v=v/(2.0d0*pi)

    bb(1,1)=(aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3))/v
    bb(2,1)=(aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3))/v
    bb(3,1)=(aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3))/v
    bb(1,2)=(aa(2,3)*aa(3,1)-aa(3,3)*aa(2,1))/v
    bb(2,2)=(aa(3,3)*aa(1,1)-aa(1,3)*aa(3,1))/v
    bb(3,2)=(aa(1,3)*aa(2,1)-aa(2,3)*aa(1,1))/v
    bb(1,3)=(aa(2,1)*aa(3,2)-aa(3,1)*aa(2,2))/v
    bb(2,3)=(aa(3,1)*aa(1,2)-aa(1,1)*aa(3,2))/v
    bb(3,3)=(aa(1,1)*aa(2,2)-aa(2,1)*aa(1,2))/v
    return
  end subroutine recp_uvec

  subroutine calchr(rho,nrxyz,va,chrg)
    implicit none
    integer,intent(in):: nrxyz
    real(kind=8),intent(in):: rho(nrxyz),va
    real(kind=8),intent(out):: chrg
    integer:: i
    chrg=0.0d0
    do i=1,nrxyz
       chrg=chrg+rho(i)
    end do
    chrg=chrg*va/nrxyz
    return
  end subroutine calchr

  subroutine calchr2(rho,nrxyz,va,chrg,achrg)
    implicit none
    integer,intent(in):: nrxyz
    real(kind=8),intent(in):: rho(nrxyz),va
    real(kind=8),intent(out):: chrg,achrg
    integer:: i
    chrg=0.0d0
    achrg=0.0d0
    do i=1,nrxyz
       chrg=chrg+rho(i)
       achrg=achrg+abs(rho(i))
    end do
    chrg=chrg*va/nrxyz
    achrg=achrg*va/nrxyz
    return
  end subroutine calchr2

  subroutine calpol(rho,nrxyz,va,pol)
    implicit none
    integer,intent(in):: nrxyz
    real(kind=8),intent(in):: rho(nrxyz,3),va
    real(kind=8),intent(out):: pol
    integer:: i
    pol = 0.0d0
    do i=1,nrxyz
       pol = pol + sqrt(rho(i,1)**2 + rho(i,2)**2 + rho(i,3)**2)
    end do
    pol = pol*va/nrxyz
    return
  end subroutine calpol

  subroutine hsort3i(n, kk, idx)
    implicit none
    integer,intent(in):: n
    integer:: kk(3,n)
    integer:: idx(n)

    integer:: i, j, ip, ic, it
    integer:: t(3)

    ! initialize idx
    do i=1,n
       idx(i) = i
    end do

    ! create the heap
    do i=2,n
       j = i
       ip = j/2
       do while ( gt3i(kk(1,j),kk(1,ip)) )
          t = kk(:,ip)
          it = idx(ip)
          kk(:,ip) = kk(:,j)
          idx(ip) = idx(j)
          kk(:,j) = t
          idx(j) = it
          j = ip
          ip = j/2
          if ( ip .eq. 0 ) exit
       end do
    end do

    ! sort it
    do i=n,1,-1
       t = kk(:,i)
       it = idx(i)
       kk(:,i) = kk(:,1)
       idx(i) = idx(1)
       kk(:,1) = t
       idx(1) = it
       j = 1
       ic = 2*j
       do while ( ic .lt. i )
          if ( ic + 1 .lt. i ) then
             if ( gt3i(kk(1,ic+1),kk(1,ic)) ) ic = ic + 1
          end if
          if ( gt3i(kk(1,ic),kk(1,j)) ) then
             t = kk(:,ic)
             it = idx(ic)
             kk(:,ic) = kk(:,j)
             idx(ic) = idx(j)
             kk(:,j) = t
             idx(j) = it
             j = ic
          else
             exit
          end if
          ic = 2*j
       end do
    end do
    return
    contains
      logical function gt3i(ka,kb)
        implicit none
        integer,intent(in):: ka(3),kb(3)

        if (ka(3).gt.kb(3)) then
           gt3i = .true.
           return
        else if (ka(3).eq.kb(3)) then
           if (ka(2).gt.kb(2)) then
              gt3i = .true.
              return
           else if (ka(2).eq.kb(2)) then
              if (ka(1).gt.kb(1)) then
                 gt3i = .true.
                 return
              end if
              gt3i = .false.
              return
           end if
           gt3i = .false.
           return
        end if
        gt3i = .false.
        return
      end function gt3i
  end subroutine hsort3i

  subroutine hanningwind(n,m,a,f,c1,c2)
    implicit none
    integer,intent(in):: n,m
    real(kind=8):: f(n,m)
    real(kind=8),intent(in):: a(n),c1,c2

    real(kind=8), parameter:: pi = 3.14159265358979323846d0
    integer:: i,j
    real(kind=8):: w

    do j=1,m
       do i=1,n
          if (a(i) .le. c1) then
             w = 1.0d0
          else if (a(i) .le. c2) then
             w = 0.5d0*( cos(pi/(c2-c1)*(a(i)-c1)) + 1.0d0 )
          else
             w = 0.0d0
          end if
          f(i,j) = w*f(i,j)
       end do
    end do
    return
  end subroutine hanningwind
  
  subroutine hsort(n, d, idx)
    implicit none
    integer,intent(in):: n
    real(kind=8):: d(n)
    integer:: idx(n)

    integer:: i, j, ip, ic, it
    real(kind=8):: t

    ! initialize idx
    do i=1,n
       idx(i) = i
    end do

    ! create the heap
    do i=2,n
       j = i
       ip = j/2
       do while ( d(j) .gt. d(ip) )
          t = d(ip)
          it = idx(ip)
          d(ip) = d(j)
          idx(ip) = idx(j)
          d(j) = t
          idx(j) = it
          j = ip
          ip = j/2
          if ( ip .eq. 0 ) exit
       end do
    end do

    ! sort it
    do i=n,1,-1
       t = d(i)
       it = idx(i)
       d(i) = d(1)
       idx(i) = idx(1)
       d(1) = t
       idx(1) = it
       j = 1
       ic = 2*j
       do while ( ic .lt. i )
          if ( ic + 1 .lt. i ) then
             if ( d(ic+1) .gt. d(ic) ) ic = ic + 1
          end if
          if ( d(ic) .gt. d(j) ) then
             t = d(ic)
             it = idx(ic)
             d(ic) = d(j)
             idx(ic) = idx(j)
             d(j) = t
             idx(j) = it
             j = ic
          else
             exit
          end if
          ic = 2*j
       end do
    end do
    return
  end subroutine hsort

  real(kind=8) function xetime(i)
    implicit none
    integer:: i
    real(kind=8),save:: tm0
    real(kind=8):: tm1
    integer(kind=8),save:: clock0
    integer(kind=8):: clock1, clockmax, clockrate, tics

    if (i.eq.0) then
       call system_clock(clock0,clockrate,clockmax)
       xetime = 0.0d0
    else if (i.eq.1) then
       call system_clock(clock1,clockrate,clockmax)
       tics = clock1-clock0
       xetime = dble(tics)/dble(clockrate)
    end if
    return
  end function xetime
    

  subroutine prntime(c,mssg)
    implicit none
    character(len=*):: c
    integer:: mssg

    if (mssg.gt.0) then
       write(6,'(1x,''TIME   '',a,''='',f20.11)') c,xetime(1)
    end if
    return
  end subroutine prntime

  subroutine chknfwrt(nf,msg,info)
    implicit none
    integer,intent(in):: nf
    character(len=*):: msg
    integer,intent(out):: info
    character(len=7):: sstat

    inquire(nf,write=sstat)
    if (sstat.eq.'NO') then
       write(6,'(1x,a)') 'Error: cannot open '//msg//' for write'
       info = 1
    end if
    return
  end subroutine chknfwrt

end module util
