!
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
!
! accelerate the convergence of input vector sequence using
! Anderson's extrapolation scheme.
!
! requires LAPACK:dposv
!
! Ref: Donald G. Anderson, J. Assoc. Computing Machinery, 12, 547 (1965)
!
! To keep the conjugation symmetry in the input vector,
! we must mix the input vectors by real numers.
! This is the reason of the definition of the inner products used here.
! 
module zextrapolate

  type zxtrapstruct
     integer:: n, m, iter
     integer:: bn, en, myrank, nproc
     complex(kind=8),dimension(:,:),pointer:: x,r
     integer,pointer:: n_count(:),n_displ(:)
  end type zxtrapstruct

contains
  subroutine zxtrap_init(in,im,inivec,xs,comm)
    use mpi
    implicit none
    integer,intent(in):: in,im,comm
    complex(kind=8):: inivec(in)
    type (zxtrapstruct) :: xs
    !
    integer:: i,err
    integer:: irnk
    !
    xs%n = in
    xs%m = im
    xs%iter = 0

    call MPI_COMM_RANK(comm,xs%myrank,err)
    call MPI_COMM_SIZE(comm,xs%nproc,err)

    allocate(xs%n_count(0:xs%nproc-1),xs%n_displ(0:xs%nproc-1))
    xs%n_count = in/xs%nproc
    do irnk=0,mod(in,xs%nproc)-1
       xs%n_count(irnk) = xs%n_count(irnk) + 1
    end do
    xs%n_displ(0) = 0
    do irnk=0,xs%nproc-2
       xs%n_displ(irnk+1) = xs%n_displ(irnk) + xs%n_count(irnk)
    end do
    xs%bn = xs%n_displ(xs%myrank)+1
    xs%en = xs%n_displ(xs%myrank)+xs%n_count(xs%myrank)

    allocate(xs%x(xs%bn:xs%en,0:im),xs%r(xs%bn:xs%en,0:im),stat=err)
    call chkalloc('zxtrap_init','alloc',err)
    do i=0,im
       xs%x(:,i) = inivec(xs%bn:xs%en)
    end do
    xs%r(:,:) = 0.0d0
  end subroutine zxtrap_init
  !
  subroutine zxtrap_done(xs)
    implicit none
    type (zxtrapstruct) :: xs
    !
    deallocate(xs%n_displ,xs%n_count)
    deallocate(xs%x,xs%r)
    return
  end subroutine zxtrap_done
  !
  subroutine zxtrap_load(nf,xs,comm)
    use mpi
    implicit none
    integer,intent(in):: nf, comm
    type (zxtrapstruct) :: xs

    integer:: i, n, m, ierr
    complex(kind=8),allocatable:: x(:),r(:)
    integer:: myrank
    
    call MPI_COMM_RANK(comm,myrank,ierr)

    if (myrank.eq.0) then
       read(nf) n, m, xs%iter
       if (xs%n .ne. n) then
          write(6,*) 'vector length of scf data is invalid'
          write(6,*) n, xs%n
          call MPI_ABORT(comm,1,ierr)
          stop
       end if
       if (xs%m .ne. m) then
          write(6,*) '# of xtrap stage is invalid'
          write(6,*) m, xs%m
          call MPI_ABORT(comm,1,ierr)
          stop
       end if
    end if
    call MPI_BCAST(xs%iter,1,MPI_INTEGER,0,comm,ierr)

    allocate(x(xs%n),r(xs%n))
    do i=0,xs%m
       if (myrank.eq.0) then
          read(nf) x, r
       end if
       call MPI_BCAST(x,xs%n,MPI_DOUBLE_COMPLEX,0,comm,ierr)
       call MPI_BCAST(r,xs%n,MPI_DOUBLE_COMPLEX,0,comm,ierr)
       xs%x(:,i) = x(xs%bn:xs%en)
       xs%r(:,i) = r(xs%bn:xs%en)
    end do
    deallocate(x,r)

    return
  end subroutine zxtrap_load
  !
  subroutine zxtrap_store(nf,xs,comm)
    use mpi
    implicit none
    integer,intent(in):: nf, comm
    type (zxtrapstruct) :: xs

    integer:: i,ierr,myrank
    complex(kind=8),allocatable:: x(:),r(:)

    call MPI_COMM_RANK(comm,myrank,ierr)
    if (myrank.eq.0) then
       write(nf) xs%n, xs%m, xs%iter
    end if
    allocate(x(xs%n),r(xs%n))
    do i=0,xs%m
       call MPI_GATHERV(xs%x(xs%bn,i),xs%n_count(xs%myrank), &
            MPI_DOUBLE_COMPLEX,x,xs%n_count,xs%n_displ, &
            MPI_DOUBLE_COMPLEX,0,comm,ierr)
       call MPI_GATHERV(xs%r(xs%bn,i),xs%n_count(xs%myrank), &
            MPI_DOUBLE_COMPLEX,r,xs%n_count,xs%n_displ, &
            MPI_DOUBLE_COMPLEX,0,comm,ierr)
       if (myrank.eq.0) then
          write(nf) x, r
       end if
    end do
    deallocate(x,r)

    return
  end subroutine zxtrap_store

  subroutine zxtrap(xs,beta,ynew,xnew,comm)
    use mpi
    use timeprof
    implicit none
    !
    type (zxtrapstruct) :: xs
    complex(kind=8),intent(in):: ynew(xs%n)
    real(kind=8),intent(in):: beta
    integer,intent(in):: comm
    !
    complex(kind=8),intent(out):: xnew(xs%n)
    !
    complex(kind=8),allocatable:: u(:),vmu(:)
    real(kind=8),allocatable:: coef(:,:),theta(:)
    complex(kind=8),allocatable:: pxnew(:)
    real(kind=8):: c(xs%m,xs%m),s(xs%m)
    complex(kind=8):: d(xs%m)
    !
    integer:: i,j,k,im,err,mz,mj,m1,m2,m3,m4,m5,m6
    complex(kind=8):: uj,vmuj

    call tpf_begin(tpf_xtrap)
    !
    ! for residual metric
    !real(kind=8):: fact
    !
    allocate(u(xs%bn:xs%en),vmu(xs%bn:xs%en),coef(xs%m,xs%m), &
         theta(xs%m),pxnew(xs%bn:xs%en),stat=err)
    call chkalloc('zxtrap','tmp',err)
    !
    mz = modulo(-xs%iter,xs%m+1)
    !
    xs%r(:,mz) = ynew(xs%bn:xs%en) - xs%x(:,mz)
    !
    ! use uniform metric
    !
    s = 0.0d0
    c = 0.0d0
    im = min(xs%m,xs%iter)
    select case (im)
    case (1)
!poption tlocal(d),psum(s,c)
       do k=xs%bn,xs%en
          m1 = modulo(mz+1,xs%m+1)
          d(1) = xs%r(k,mz) - xs%r(k,m1)
          s(1) = s(1) + dble(xs%r(k,mz))*dble(d(1)) &
                  + aimag(xs%r(k,mz))*aimag(d(1))

          c(1,1) = c(1,1) + dble(d(1))*dble(d(1)) &
               + aimag(d(1))*aimag(d(1))
       end do
    case (2)
!$omp parallel do private(d,i,j,mj),reduction(+:s,c)
!poption tlocal(d),psum(s,c)
       do k=xs%bn,xs%en
          do j=1,2
             mj = modulo(mz+j,xs%m+1)
             d(j) = xs%r(k,mz) - xs%r(k,mj)
             s(j) = s(j) + dble(xs%r(k,mz))*dble(d(j)) &
                  + aimag(xs%r(k,mz))*aimag(d(j))
          end do
          do j=1,2
             do i=j,2
                c(i,j) = c(i,j) + dble(d(i))*dble(d(j)) &
                     + aimag(d(i))*aimag(d(j))
             end do
          end do
       end do
    case (3)
!$omp parallel do private(d,i,j,mj),reduction(+:s,c)
!poption tlocal(d),psum(s,c)
       do k=xs%bn,xs%en
          do j=1,3
             mj = modulo(mz+j,xs%m+1)
             d(j) = xs%r(k,mz) - xs%r(k,mj)
             s(j) = s(j) + dble(xs%r(k,mz))*dble(d(j)) &
                  + aimag(xs%r(k,mz))*aimag(d(j))
          end do
          do j=1,3
             do i=j,3
                c(i,j) = c(i,j) + dble(d(i))*dble(d(j)) &
                     + aimag(d(i))*aimag(d(j))
             end do
          end do
       end do
    case (4)
!$omp parallel do private(d,i,j,mj),reduction(+:s,c)
!poption tlocal(d),psum(s,c)
       do k=xs%bn,xs%en
          do j=1,4
             mj = modulo(mz+j,xs%m+1)
             d(j) = xs%r(k,mz) - xs%r(k,mj)
             s(j) = s(j) + dble(xs%r(k,mz))*dble(d(j)) &
                  + aimag(xs%r(k,mz))*aimag(d(j))
          end do
          do j=1,4
             do i=j,4
                c(i,j) = c(i,j) + dble(d(i))*dble(d(j)) &
                     + aimag(d(i))*aimag(d(j))
             end do
          end do
       end do
    case (5)
!$omp parallel do private(d,i,j,mj),reduction(+:s,c)
!poption tlocal(d),psum(s,c)
       do k=xs%bn,xs%en
          do j=1,5
             mj = modulo(mz+j,xs%m+1)
             d(j) = xs%r(k,mz) - xs%r(k,mj)
             s(j) = s(j) + dble(xs%r(k,mz))*dble(d(j)) &
                  + aimag(xs%r(k,mz))*aimag(d(j))
          end do
          do j=1,5
             do i=j,5
                c(i,j) = c(i,j) + dble(d(i))*dble(d(j)) &
                     + aimag(d(i))*aimag(d(j))
             end do
          end do
       end do
    case (6)
!$omp parallel do private(d,i,j,mj),reduction(+:s,c)
!poption tlocal(d),psum(s,c)
       do k=xs%bn,xs%en
          do j=1,6
             mj = modulo(mz+j,xs%m+1)
             d(j) = xs%r(k,mz) - xs%r(k,mj)
             s(j) = s(j) + dble(xs%r(k,mz))*dble(d(j)) &
                  + aimag(xs%r(k,mz))*aimag(d(j))
          end do
          do j=1,6
             do i=j,6
                c(i,j) = c(i,j) + dble(d(i))*dble(d(j)) &
                     + aimag(d(i))*aimag(d(j))
             end do
          end do
       end do
    case default
!$omp parallel do private(d,i,j,mj),reduction(+:s,c)
!poption tlocal(d),psum(s,c)
       do k=xs%bn,xs%en
          do j=1,im
             mj = modulo(mz+j,xs%m+1)
             d(j) = xs%r(k,mz) - xs%r(k,mj)
             s(j) = s(j) + dble(xs%r(k,mz))*dble(d(j)) &
                  + aimag(xs%r(k,mz))*aimag(d(j))
          end do
          do j=1,im
             do i=j,im
                c(i,j) = c(i,j) + dble(d(i))*dble(d(j)) &
                     + aimag(d(i))*aimag(d(j))
             end do
          end do
       end do
    end select
    call tpf_begin(tpf_mpi)
    call MPI_ALLREDUCE(s,theta,xs%m,MPI_DOUBLE_PRECISION,MPI_SUM,comm,err)
    call MPI_ALLREDUCE(c,coef,xs%m*xs%m,MPI_DOUBLE_PRECISION,MPI_SUM,comm,err)
    call tpf_end(tpf_mpi)
    !
    if (im .gt. 0) then
       call dposv('L',im,1,coef,xs%m,theta,xs%m,err)
       if (err .lt. 0) then
          write(6,*)'zxtrap: dposv: invalid arg in ',err
          stop
       else if (err .gt. 0) then
          write(6,*) 'zxtrap: input vector sequence become linealy dependent'
          write(6,*) 'zxtrap: only last vector used'
          im = 0
       end if
    end if
    !
    ! theta has got
    !
    select case (im)
    case (1)
       m1 = modulo(mz+1,xs%m+1)
       do j=xs%bn,xs%en
          uj = xs%x(j,mz) + theta(1)*(xs%x(j,m1) - xs%x(j,mz))
          vmuj = xs%r(j,mz) + theta(1)*(xs%r(j,m1) - xs%r(j,mz))

          pxnew(j) = uj + beta * vmuj
       end do
    case (2)
       m1 = modulo(mz+1,xs%m+1)
       m2 = modulo(mz+2,xs%m+1)
       do j=xs%bn,xs%en
          uj = xs%x(j,mz) + theta(1)*(xs%x(j,m1) - xs%x(j,mz)) &
               + theta(2)*(xs%x(j,m2) - xs%x(j,mz))
          vmuj = xs%r(j,mz) + theta(1)*(xs%r(j,m1) - xs%r(j,mz)) &
               + theta(2)*(xs%r(j,m2) - xs%r(j,mz))

          pxnew(j) = uj + beta * vmuj
       end do
    case (3)
       m1 = modulo(mz+1,xs%m+1)
       m2 = modulo(mz+2,xs%m+1)
       m3 = modulo(mz+3,xs%m+1)
       do j=xs%bn,xs%en
          uj = xs%x(j,mz) + theta(1)*(xs%x(j,m1) - xs%x(j,mz)) &
               + theta(2)*(xs%x(j,m2) - xs%x(j,mz)) &
               + theta(3)*(xs%x(j,m3) - xs%x(j,mz))
          vmuj = xs%r(j,mz) + theta(1)*(xs%r(j,m1) - xs%r(j,mz)) &
               + theta(2)*(xs%r(j,m2) - xs%r(j,mz)) &
               + theta(3)*(xs%r(j,m3) - xs%r(j,mz))

          pxnew(j) = uj + beta * vmuj
       end do
    case (4)
       m1 = modulo(mz+1,xs%m+1)
       m2 = modulo(mz+2,xs%m+1)
       m3 = modulo(mz+3,xs%m+1)
       m4 = modulo(mz+4,xs%m+1)
       do j=xs%bn,xs%en
          uj = xs%x(j,mz) + theta(1)*(xs%x(j,m1) - xs%x(j,mz)) &
               + theta(2)*(xs%x(j,m2) - xs%x(j,mz)) &
               + theta(3)*(xs%x(j,m3) - xs%x(j,mz)) &
               + theta(4)*(xs%x(j,m4) - xs%x(j,mz))
          vmuj = xs%r(j,mz) + theta(1)*(xs%r(j,m1) - xs%r(j,mz)) &
               + theta(2)*(xs%r(j,m2) - xs%r(j,mz)) &
               + theta(3)*(xs%r(j,m3) - xs%r(j,mz)) &
               + theta(4)*(xs%r(j,m4) - xs%r(j,mz))

          pxnew(j) = uj + beta * vmuj
       end do
    case (5)
       m1 = modulo(mz+1,xs%m+1)
       m2 = modulo(mz+2,xs%m+1)
       m3 = modulo(mz+3,xs%m+1)
       m4 = modulo(mz+4,xs%m+1)
       m5 = modulo(mz+5,xs%m+1)
       do j=xs%bn,xs%en
          uj = xs%x(j,mz) + theta(1)*(xs%x(j,m1) - xs%x(j,mz)) &
               + theta(2)*(xs%x(j,m2) - xs%x(j,mz)) &
               + theta(3)*(xs%x(j,m3) - xs%x(j,mz)) &
               + theta(4)*(xs%x(j,m4) - xs%x(j,mz)) &
               + theta(5)*(xs%x(j,m5) - xs%x(j,mz))
          vmuj = xs%r(j,mz) + theta(1)*(xs%r(j,m1) - xs%r(j,mz)) &
               + theta(2)*(xs%r(j,m2) - xs%r(j,mz)) &
               + theta(3)*(xs%r(j,m3) - xs%r(j,mz)) &
               + theta(4)*(xs%r(j,m4) - xs%r(j,mz)) &
               + theta(5)*(xs%r(j,m5) - xs%r(j,mz))

          pxnew(j) = uj + beta * vmuj
       end do
    case (6)
       m1 = modulo(mz+1,xs%m+1)
       m2 = modulo(mz+2,xs%m+1)
       m3 = modulo(mz+3,xs%m+1)
       m4 = modulo(mz+4,xs%m+1)
       m5 = modulo(mz+5,xs%m+1)
       m6 = modulo(mz+6,xs%m+1)
       do j=xs%bn,xs%en
          uj = xs%x(j,mz) + theta(1)*(xs%x(j,m1) - xs%x(j,mz)) &
               + theta(2)*(xs%x(j,m2) - xs%x(j,mz)) &
               + theta(3)*(xs%x(j,m3) - xs%x(j,mz)) &
               + theta(4)*(xs%x(j,m4) - xs%x(j,mz)) &
               + theta(5)*(xs%x(j,m5) - xs%x(j,mz)) &
               + theta(6)*(xs%x(j,m6) - xs%x(j,mz))
          vmuj = xs%r(j,mz) + theta(1)*(xs%r(j,m1) - xs%r(j,mz)) &
               + theta(2)*(xs%r(j,m2) - xs%r(j,mz)) &
               + theta(3)*(xs%r(j,m3) - xs%r(j,mz)) &
               + theta(4)*(xs%r(j,m4) - xs%r(j,mz)) &
               + theta(5)*(xs%r(j,m5) - xs%r(j,mz)) &
               + theta(6)*(xs%r(j,m6) - xs%r(j,mz))

          pxnew(j) = uj + beta * vmuj
       end do
    case default
       u(:) = xs%x(:,mz)
       vmu(:) = xs%r(:,mz)
       do j=1,im
          mj = modulo(mz+j,xs%m+1)
          u(:) = u(:) + theta(j)*(xs%x(:,mj) - xs%x(:,mz))
          vmu(:) = vmu(:) + theta(j)*(xs%r(:,mj) - xs%r(:,mz))
       end do

       pxnew(:) = u(:) + beta * vmu(:)
    end select

    mz = modulo(mz-1,xs%m+1)
    xs%x(:,mz) = pxnew
    xs%iter = xs%iter + 1

    call tpf_begin(tpf_mpi)
    call MPI_ALLGATHERV(pxnew,xs%n_count(xs%myrank),MPI_DOUBLE_COMPLEX, &
         xnew,xs%n_count,xs%n_displ,MPI_DOUBLE_COMPLEX,comm,err)
    call tpf_end(tpf_mpi)
    !
    deallocate(u,vmu,coef,theta,pxnew)

    call tpf_end(tpf_xtrap)
    return
  end subroutine zxtrap
end module zextrapolate
